#include <iostream>
#include <fstream>
#include <sstream>

#include <unistd.h>

#include "midic.hpp"
#include "midip.hpp"
#include "option.h"

namespace OPTION
{
	bool enable_pan = false;
	bool enable_exp = false;
	bool enable_vol = false;
	bool enable_vel = false;
	bool enable_mod = false;
	bool enable_hold = true;
	bool enable_pitch = false;
	bool enable_quantize = false;
	bool enable_octave_rev = false;
	int max_length = 1920;
}

//==============================================================================

int main(int argc, char **argv)
{
	int result;
	while( (result = getopt(argc,argv,"pbmlevohq:") ) !=-1 )
	{
		switch(result){
			case 'p': OPTION::enable_pan = true; break;
			case 'b': OPTION::enable_pitch = true; break;
			case 'm': OPTION::enable_mod = true; break;
			case 'l': OPTION::enable_vol = true; break;
			case 'e': OPTION::enable_exp = true; break;
			case 'v': OPTION::enable_vel = true; break;
			case 'o': OPTION::enable_octave_rev = true; break;
			case 'h': OPTION::enable_hold = false; break;

			case 'q':
			{
				OPTION::enable_quantize = true;
				int num = atoi(optarg);
				if (num == 0 || num > 192) 
				{
					std::cerr << "error: -q[n] / n requires 1 ~ 192\n" << std::endl;
					return -1;
				}
				OPTION::max_length = 1920 / num;
				break;
			}
		}
	}

	if (optind == argc) {
		std::cerr << "usage: " << argv[0] << " [option] input.mid [output.mml]" << std::endl;
		std::cerr << std::endl;
		std::cerr << "-v   : enable velocity" << std::endl;
		std::cerr << "-e   : enable expression" << std::endl;
		std::cerr << "-l   : enable level change" << std::endl;
		std::cerr << "-b   : enable pitch bend" << std::endl;
		std::cerr << "-p   : enable panpot" << std::endl;
		std::cerr << "-o   : reverse octave shift" << std::endl;
		std::cerr << "-h   : disable hold" << std::endl;
		std::cerr << "-q n : enable quantize (n requires mml length 1 ~ 192)" << std::endl;
		return 0;
	}

	{
		std::ifstream filein;
		filein.open( argv[optind] , std::ios::in | std::ios::binary);
		
		if (filein) {
			char *p;

			filein.seekg(0, std::ios::end);
			unsigned int sz = filein.tellg();
			filein.seekg(0, std::ios::beg);

			p = new char[sz];
			filein.read( p, sz );
			if (filein.bad()) {
				std::cerr << "File read error, abort..." << std::endl;
				return -1;
			}
			filein.close();

			MIDI::MIDIC *midic = new MIDI::MIDIC;
			MIDI::DMIDI *dm = midic->convert_smf(p);
			if (!dm) {
				return 1;
			}

			MIDI::MIDIP *midip = new MIDI::MIDIP;
			midip->SetConvertedSmf(dm);

			std::ostringstream oss;
			midip->Render(oss);

			if ((optind + 2) == argc)
			{
				std::ofstream ofs(argv[optind + 1]);
				if (ofs)
				{
					ofs << oss.str() << std::endl;
					ofs.close();
				}
				else
				{
					std::cerr << "File" << argv[optind + 1] << "open error." << std::endl;
				}
			}
			else
			{
				std::cout << oss.str() << std::endl;
			}

			delete midip;
			delete midic;
		} else {
			std::cerr << "File open error, abort..." << std::endl;
			return -1;
		}
	}

	return 0;
}
