#include <sstream>
#include <functional>
#include "midi.hpp"
#include "midip.hpp"
#include "mmlgen.h"
#include "option.h"

namespace MIDI {

//--------------------------------------
// 
//--------------------------------------
void 
MIDIP::SetConvertedSmf(DMIDI *dmidi)
{
	this->dmidi = dmidi;
	eventPoint = dmidi->events.begin();
	endPoint = dmidi->events.end();
	loopPoint = dmidi->events.end();
	loopCount = 0;
}

void 
MIDIP::AnalyzeControlChange(std::vector<DMIDI_EVENT>::iterator& e)
{
	unsigned int ch = e->ch;
	unsigned int lsb = e->lsb;
	unsigned int msb = e->msb;
	switch (lsb) {
		case HOLD: 
		{
			if (OPTION::enable_hold)
			{
				track[ch].hold = msb;
				if (msb == 0)
				{
					// z[hĂ锭
					if (track[ch].notes.size())
					{
						std::list<note_>::iterator it = track[ch].notes.begin();
						while ( it != track[ch].notes.end() )
						{
							if (it->hold != 0)
							{
								DMIDI_EVENT d = *e;
								d.mes = NOTE_OFF;
								d.lsb = it->note;
								mod->add(it->ch, d);
								it = track[ch].notes.erase( it );
							}
							else
							{
								++it;
							}
						}
					}
				}
			}
			break;
		}

		case MODULATION:
		{
			if (!OPTION::enable_mod) break;
			break;
		}

		case PANPOT:
		{
			if (OPTION::enable_pan) 
			{
				DMIDI_EVENT d = *e;
				d.mes = CC;
				d.lsb = PANPOT;
				d.msb = msb;
				mod->add(d);
				break;
			}
			break;
		}
		
		case VOLUME: 
		case EXPRESSION: 
		{
			if (lsb == VOLUME) 
			{
				if (!OPTION::enable_vol) break;
				track[ch].vol = msb;
			}
			else
			{
				if (!OPTION::enable_exp) break;
				track[ch].exp = msb;
			}
			int v = static_cast<int>(
						static_cast<double>(track[ch].vol) 
						* static_cast<double>(track[ch].exp) 
						/ 127.0L
					);
			DMIDI_EVENT d = *e;
			d.mes = CC;
			d.lsb = VOLUME;
			d.msb = v;
			mod->add(d);
			break;
		}

		case LOOP_POINT: 
		{
			for (int i = 0; i < 16; i++)
			{
				DMIDI_EVENT d = *e;
				d.mes = CC;
				d.ch = i;
				d.lsb = LOOP_POINT;
				d.msb = 0;
				mod->add(d);
			}
			break;
		}

		case DATAENTRY_MSB: 
		{
			if (track[ch].rpn_lsb == 0 && track[ch].rpn_msb == 0) 
			{
				if (msb > 32) msb = 0;
				track[ch].bendrange = msb;
			}
			break;
		}
		case RPN_LSB: 
		{
			track[ch].rpn_lsb = msb;
			break;
		}
		case RPN_MSB: 
		{
			track[ch].rpn_msb = msb;
			break;
		}

		case VOICE_RESERVE: 
		{
			if (msb < 1) 
				msb = 1;

			track[ch].voice_reserve = msb;
			mod->SetReserve(ch, msb);
			break;
		}
		default: 
		{
			//mod->add(*e);
			break;
		}
	}
}

//--------------------------------------
// ret: none
//--------------------------------------
void 
MIDIP::Analyse(std::vector<DMIDI_EVENT>::iterator& e)
{
	const unsigned int ch = e->ch;
	const unsigned int mes = e->mes;
	const unsigned int lsb = e->lsb;
	const unsigned int msb = e->msb;

	// AiCY
	switch (mes & 0xf0) 
	{
		case DMIDI_SYS: 
		{
			switch (ch | DMIDI_SYS) 
			{
				case DMIDI_SYS_BPM: 
				{
					DMIDI_EVENT d = *e;
					d.mes = DMIDI_SYS_BPM;
					d.ch = 0;
					mod->add(d);
					break;
				}
				case DMIDI_SYS_ENDDATA:
				{
					for (int i = 0; i < 16; i++) 
					{
						DMIDI_EVENT d = *e;
						d.mes = DMIDI_SYS_ENDDATA;
						d.ch = i;
						mod->add(d);
					}
					break;
				}
				default: 
				{
					break;
				}
			}
			break;
		}

		case CC: 
		{
			AnalyzeControlChange(e);
			break;
		}
		
		case PC: 
		{
			mod->add(*e);
			break;
		}
		
		case PITCHBEND: 
		{
			if (OPTION::enable_pitch) 
			{
				int pitch = static_cast<short>(lsb | msb << 8);
				pitch = static_cast<int>(
					static_cast<double>(pitch) * static_cast<double>(track[ch].bendrange) / 12.0L / 10.6L
				);
				DMIDI_EVENT d = *e;
				d.mes = PITCHBEND;
				d.lsb = pitch & 0xff;
				d.msb = (pitch >> 8) & 0xff;
				mod->add(d);
			}
			break;
		}

		case NOTE_ON: 
		{
			if (0 == msb) 
			{
				//printf("ARIENAI YO!\n");
			}
			else 
			{
				int vel = msb;
				{
					int virtual_ch = ch;
					// U[u`lSĖ܂Ăꍇ
					// ^C~OƂ̂
					if (track[ch].notes.size() == track[ch].voice_reserve)
					{
						unsigned int clock = 0xffffffffL;
						std::list<note_>::iterator it, erase_it;
						it = track[ch].notes.begin();
						for ( ; it != track[ch].notes.end(); ++it)
						{
							if (clock >= it->clock)
							{
								clock = it->clock;
								virtual_ch = it->ch;
								erase_it = it;
							}
						}
						if (track[ch].voice_reserve > 1)
						{
							DMIDI_EVENT d = *e;
							d.mes = NOTE_OFF;
							d.lsb = erase_it->note;
							mod->add(virtual_ch, d);
						}
						track[ch].notes.erase( erase_it );
					}
					else
					{
						// 󂫃`lT
						unsigned int empty_ch = 0;
						for ( ; empty_ch < track[ch].voice_reserve; ++empty_ch)
						{
							int used = 0;
							std::list<note_>::iterator it = track[ch].notes.begin();
							for ( ; it != track[ch].notes.end(); ++it)
							{
								if (it->ch == empty_ch) 
								{
									used = 1;
									break;
								}
							}
							if (!used) 
							{
								break;
							}
						}

						virtual_ch = empty_ch;
					}
					note_ n;
					n.ch = virtual_ch;
					n.note = lsb;
					n.vel = vel;
					n.clock = e->clock;
					n.hold = 0;
					track[ch].notes.push_back(n);

					mod->add(virtual_ch, *e);
				}
				track[ch].key_on = true;
				track[ch].note = lsb;
				track[ch].velocity = vel;
			}
			break;
		}

		case NOTE_OFF: 
		{
			if (track[ch].notes.size())
			{
				std::list<note_>::iterator it, erase_it;
				it = track[ch].notes.begin();
				while ( it != track[ch].notes.end() )
				{
					if (it->note == lsb)
					{
						if (track[ch].hold)
						{
							it->hold = 1;
							++it;
						}
						else
						{
							DMIDI_EVENT d = *e;
							d.mes = NOTE_OFF;
							d.lsb = it->note;
							mod->add(it->ch, d);
							it = track[ch].notes.erase( it );
						}
					}
					else
					{
						++it;
					}
				}
			}
			else
			{
				//printf("NoteOff ignored\n");
			}
			break;
		}
		
		default: 
		{
			break;
		}
	}
}


//--------------------------------------
// ret: none
//--------------------------------------
void 
MIDIP::Render(std::ostringstream& str)
{

	std::vector<DMIDI_EVENT>::iterator it = dmidi->events.begin();
	std::vector<DMIDI_EVENT>::iterator end = dmidi->events.end();

	if (OPTION::enable_quantize)
	{
		unsigned int division = OPTION::max_length;
		// GcɃNI^CY
		while (it != end) {
			unsigned int mod = it->clock % division;
			if (mod >= division /2) 
			{
				it->clock += division - mod;
			} 
			else
			{
				it->clock -= mod;
			}
			it++;
		}
	}

	it = dmidi->events.begin();
	while (it != end)
	{
		Analyse(it);
		it++;
	}

	if (!OPTION::enable_octave_rev)
	{
		str << "#REV;\n";
	}
	for (unsigned int i = 0; i < mod->GetSize(); ++i)
	{
		bool ch_header = false;
		Tracks& trk = mod->v_ch[i];
		for (unsigned int j = 0; j < trk.GetSize(); j++)
		{
			DMIDI& dm = trk.v_track[j];
			if (dm.events.size() > 2) // [vƃGh}[ÑgbN𖳎邪AQ邩
			{
				if (!ch_header)
				{
					str << "//=== Channel " << (i + 1) << " =====================" << std::endl;
					ch_header = true;
				}
				str << "// ch:" << i+1 << " pt:" << j+1 << std::endl;
				mmlgen(str, &dm);
			}
		}
	}

}


} // namespace MIDI
