#ifndef MIDIP_HPP
#define MIDIP_HPP

#include <sstream>
#include <iostream>
#include <vector>
#include <list>

#include "midid.hpp"

namespace MIDI
{
static const unsigned int BASE_TEMPO = 120;

struct Tracks
{
	std::vector<DMIDI> v_track;
	int reserve;
	Tracks() {
		v_track.resize(1);
		reserve = 1;
	}
	size_t GetSize()
	{
		return v_track.size();
	}
	void SetReserve(int r)
	{
		reserve = r;
		if (v_track.size() < static_cast<unsigned int>(r))
		{
			v_track.resize(r);
		}
	}

	void add(DMIDI_EVENT& event) {
		int channel = reserve;
		if (event.mes == DMIDI_SYS_ENDDATA)
			channel = GetSize();
		for (int i = 0; i < channel; i++)
		{
			v_track[i].events.push_back(event);
		}
	}
	void add(unsigned int track, DMIDI_EVENT& event) {
		v_track[track].events.push_back(event);
	}
};

struct Modules
{
	std::vector<Tracks> v_ch;
	Modules() {
		v_ch.resize(32);
	}
	
	size_t GetSize()
	{
		return v_ch.size();
	}

	void SetReserve(int ch, int r)
	{
		v_ch[ch].SetReserve(r);
	}
	void add(DMIDI_EVENT& event) {
		v_ch[event.ch].add(event);
	}
	void add(unsigned int track, DMIDI_EVENT& event) {
		v_ch[event.ch].add(track, event);
	}
	void add(unsigned int ch, unsigned int track, DMIDI_EVENT& event) {
		v_ch[ch].add(track, event);
	}
};

class MIDIP
{
private:
	Modules *mod;
	DMIDI *dmidi;
	std::vector<DMIDI_EVENT>::iterator eventPoint;
	std::vector<DMIDI_EVENT>::iterator endPoint;
	std::vector<DMIDI_EVENT>::iterator loopPoint;

	unsigned int clockCount;
	unsigned int loopCount;
	unsigned int sampleRate;
	unsigned int tempo;

	void AnalyzeControlChange(std::vector<DMIDI_EVENT>::iterator& e);

	struct note_ {
		unsigned int clock;
		unsigned int ch;
		unsigned int note;
		unsigned int vel;
		unsigned int hold;
	};
	
	struct TrackStatus {
		unsigned char prog;
		unsigned char vol;
		unsigned char exp;
		unsigned char pan;

		unsigned char bendrange;
		int pitchbend;
		int porta_time;
		int porta_switch;

		int modulation;
		int mod_depth;
		int mod_speed;
		int mod_count;

		unsigned int rpn_msb;
		unsigned int rpn_lsb;

		bool key_on;
		unsigned int note;
		unsigned int velocity;
		unsigned int voice_reserve;
		int hold;

		std::list<note_> notes;

		TrackStatus() {
			key_on = false;
			exp = 127;
			vol = 127;
			note = 0;
			bendrange = 12;
			rpn_msb = 0;
			rpn_lsb = 0;
			voice_reserve = 1;
			hold = 0;
		}
	} track[32];

public:
	void SetConvertedSmf(DMIDI *dmidi);
	void Analyse(std::vector<DMIDI_EVENT>::iterator& e);
	void Render(std::ostringstream& str);

	MIDIP() {
		tempo = BASE_TEMPO;
		mod = new Modules;
	}
	~MIDIP() {
		if (mod)
		{
			delete mod;
		}
	}
};

}

#endif
